.TH  "TCLSCAN" N
.SH NAME
tclscan \- scan tcl script(s) for definitions of non-buildin tcl
commands and for global variables and store the result in a *.ini file

.SH SYNOPSIS
\fBtclscan\fP \fIfile1.tcl\fP ?\fIfile2.tcl\fP? ...

.SH DESCRIPTION
The \fItclscan\fP program is used to scan a tcl script for new command
definitions and global variables. Each file that is given as argument
will be scanned separately.

The only command definitions the program will recognize are the
definitions of procedures (using \fIproc\fP) and images (using \fIimage
create\fP).

Like in Tcl, a new command definition with the same name will
overwrite an existing definition. Since this is not the case for
widgets these are recognized directly when parsing with \fItclparse\fP.
Global variables are found in declarations with \fIglobal\fP in
command bodies or in definitions with the tcl commands \fIappend\fP,
\fIcatch\fP, \fIfile\fP, \fIgets\fP, \fIlappend\fP, \fIregexp\fP,
\fIregsub\fP, \fIscan\fP and \fIupvar\fP outside of command bodies.

The scanned informations are stored in an \fIinformation file\fP. The
name of this \fIinformation file\fP is build by appending the suffix
\fB.ini\fP to the name of the file that is currently being scanned.

This \fIinformation file\fP is \fBused\fP by the \fItclparse\fP
program which must know the new commands and global variables.

If \fItclscan\fP encounters the \fIsource\fP command, it will try to
open the associated file. However, this will only work if the argument
of source is a real filename (absolute or relative doesn't
matter) and no substitutions are used. If necessary,
the concerning file has to be scanned manually with \fItclscan\fP and
merged to the information file.

Several \fIinformation files\fP can be merged with the \fItclmerge\fP
program. For more information see the manual page for \fItclmerge\fP.


.SH BUGS AND LIMITATIONS
\fItclscan\fP was written with the GNU project C++ Compiler on an intel-based
linux-pc. Therefore we don't know how the program will react in
other environments.

The program only looks for command names after newline
characters. Therefore it sometimes finds a procedure or variable
declaration that is none. Example:

switch -- $variable {
  a -
  b -
  c { puts "a, b or c" }

  proc { puts "proc" }

  d -
  e -
  f { puts "d, e or f" }

  default { puts "garbage" }
}

\fItclscan\fP will find the option 'proc' to be a procedure
declaration. The following body will be used as procedure name and
the following option 'd' as the corresponding command body.

This behaviour may occur with other commands. Normally this should not
interfere with the other information, since the procedure name from
above isn't likely to be a valid procedure name of a tcl program. 

In a situation like described above the program might tell you a
\fBsegmentation fault\fP if an extra large body is being taken for a
procedure name.

.SH AUTHORS
Stephan Weiss & Heiko Grossman

Homepage: http://www.informatik.uni-stuttgart.de/ipvr/swlab/sopra/tclsyntax/tclparseHome.html

.SH VERSION 
V2.00 from 14. March 1997 

This program is distributed in the hope that it will be useful, but
\fBWITHOUT ANY WARRANTY\fP; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

.SH RELATED COMMANDS
\fItclscan\fP is part of a package which contains also
the programs \fItclmerge\fP and \fItclparse\fP. See the corresponding
manual pages for details.






