/*********************************************************************/
/* tclparse V2.00 from 14.March 1997                                 */
/* Copyright 1996/97 Stephan Weiss, Heiko Grossmann                  */
/*                   University of Stuttgart, Germany                */
/* For details see the copyright file included in the distribution.  */
/* This notice must not be removed from the file.                    */
/*                                                                   */
/* routines for managing command and variable environments           */
/*                                                                   */
/*********************************************************************/

#include "../include/const.h"
#include "../include/handler.h"
#include "../include/parsing.h"
#include "../include/display.h"
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdlib.h>


//######################################################################


template<class TYPE> CList<TYPE>::CList( void )
{
  root = new SElement;
  root->prev = root;
  root->next = root;
  internalPtr = root->next;
}


template<class TYPE> CList<TYPE>::~CList( void )
{
  //Ein Zeiger definieren, der die gesamte Liste durchlaeuft.
  SElement *dummy;

  //Den Zeiger auf das erste Element setzen.
  dummy = root->next;

  //Die Liste solange durchlaufen, bis der Zeiger wieder auf der Wurzel steht.
  //Existieren keine Elemente, wird die Schleife nicht durchlaufen.
  while( dummy != root )
    {
      //Den Zeiger auf das naechste Element setzen.
      dummy = dummy->next;

      //Das vorherige Element loeschen.
      delete dummy->prev;
    }

  //Zum Schluss noch den Wurzelzeiger loeschen.
  delete root;
}


template<class TYPE> TYPE *CList<TYPE>::First( void )
{
  if( root->next == root )
    return NULL;

  internalPtr = root->next;
  return &internalPtr->content;
}


template<class TYPE> TYPE *CList<TYPE>::Next( void )
{
  if( internalPtr->next == root )
    return NULL;

  internalPtr = internalPtr->next;
  return &internalPtr->content;
}


template<class TYPE> void CList<TYPE>::Insert( TYPE &element )
{
  //Ein neues Element erzeugen.
  SElement *newElement = new SElement;

  //Das neue Element initialisieren.
  newElement->content = element;

  //Einen Hilfszeiger definieren und ihn auf das letzte Element setzen.
  SElement *last;
  last = root->prev;

  //Das Element wird hinten angehaengt.
  last->next = newElement;
  newElement->prev = last;

  newElement->next = root;
  root->prev = newElement;
}


template<class TYPE> int CList<TYPE>::Remove( TYPE &element )
{
  //Ist die Liste leer, kann nichts entfernt werden.
  if( IsEmpty() )
    return 0;

  //Einen Hilfszeiger auf das erste Element setzen.
  SElement *dummy = root->next;

  //Zwei weitere Hilfszeiger deklarieren, die beim Loeschen benoetigt werden.
  SElement *prev, *next;

  //Dieser Wert gibt nach der Bearbeitung Auskunft darueber, ob mindestens
  //ein Element geloescht wurde oder nicht.
  int deleted = 0;

  //Die Liste durchlaufen und alle passenden Elemente entfernen.
  while( dummy != root )
    {
      //Das aktuelle Element pruefen.
      if( dummy->content == element )
	{
	  //Das vorherige und das naechste Element muessen miteinander
	  //verknuepft werden.
	  prev = dummy->prev;
	  next = dummy->next;

	  prev->next = next;
	  next->prev = prev;

	  //Das Element loeschen.
	  delete dummy;

	  //Die Loeschaktion vermerken;
	  deleted = 1;
	}

      dummy = dummy->next;
    }

  return deleted;
}


template<class TYPE> TYPE *CList<TYPE>::Get( TYPE &element )
{
  //Ist die Liste leer, kann nichts gefunden werden.
  if( IsEmpty() )
    return NULL;

  //Einen Hilfszeiger auf das erste Element setzen.
  SElement *dummy = root->next;

  //Die Liste durchlaufen.
  while( dummy != root )
    {
      //Das aktuelle Element pruefen.
      if( dummy->content == element )
	return &dummy->content;

      else
	dummy = dummy->next;
    }

  return NULL;
}


template<class TYPE> int CList<TYPE>::Contains( TYPE &element )
{
  //Einen Hilfszeiger definieren und ihn auf das erste Element setzen.
  SElement *dummy;
  dummy = root->next;

  //Alle Elemente ausser dem Wurzelelement testen.
  while( dummy != root )
    {
      //Den Vergleich durchfuehren.
      if( dummy->content == element )
	return 1;
      else
	dummy = dummy->next;
    }

  //Kommt die Programmausfuehrung bis hier, so wurde das Element nicht gefunden.
  return 0;
}


template<class TYPE> int CList<TYPE>::IsEmpty( void )
{
  if( root->next == root )
    return 1;

  else
    return 0;
}


template<class TYPE> CList<TYPE> &CList<TYPE>::operator=( CList &list )
{
  if( this == &list )
    return *this;

  //Zuerst die vorhandene Liste loeschen. Dazu einen Zeiger definieren,
  //der die gesamte Liste durchlaeuft.
  SElement *dummy;

  //Den Zeiger auf das erste Element setzen.
  dummy = root->next;

  //Die Liste solange durchlaufen, bis der Zeiger wieder auf der Wurzel steht.
  //Existieren keine Elemente, wird die Schleife nicht durchlaufen.
  while( dummy != root )
    {
      //Den Zeiger auf das naechste Element setzen.
      dummy = dummy->next;

      //Das vorherige Element loeschen.
      delete dummy->prev;
    }

  root->next = root;
  root->prev = root;
  internalPtr = root->next;

  //Dann alle Elemente der angegebenen Liste hinzufuegen.
  TYPE *content = list.First();

  while( content != NULL )
    {
      Insert( *content );
      content = list.Next();
    }
  return *this;
}


//######################################################################


CTclDataType::CTclDataType( void )
{
  name = NULL;
  calls = 0;
}


CTclDataType::CTclDataType( char *newName, int newCalls=0 )
{
  name = new char[ strlen( newName )+1 ];
  strcpy( name, newName );
  calls = newCalls;
}


CTclDataType::CTclDataType( const CTclDataType &dataType )
{
  delete name;
  name = new char[ strlen( dataType.name )+1 ];
  strcpy( name, dataType.name );
  calls = dataType.calls;
}


CTclDataType::~CTclDataType( void )
{
  delete name;
}


char *CTclDataType::Name( void )
{
  CopyToBuffer( name );
  return Buffer();
}


int CTclDataType::Calls( void )
{
  return calls;
}


void CTclDataType::IncCalls( void )
{
  calls++;
}


void CTclDataType::SetName( char *newName )
{
  delete name;
  name = new char[ strlen( newName )+1 ];
  strcpy( name, newName );
}


void CTclDataType::SetCalls( int newCalls )
{
  calls = newCalls;
}


int CTclDataType::operator==( CTclDataType &dataType )
{
  //Zum Vergleich wird nur der Name verwendet.
  if( !strcmp( name, dataType.name ) )
    return 1;

  else
    return 0;
}


CTclDataType &CTclDataType::operator=( CTclDataType &dataType )
{
  //Bei Selbstzuweisung muss man nichts machen.
  if( &dataType == this )
    return *this;

  delete name;
  name = new char[ strlen( dataType.name )+1 ];
  strcpy( name, dataType.name );
  calls = dataType.calls;
  return *this;
}


//######################################################################


CCommand::CCommand( void )
{
  commandID = NULL;
  list = NULL;
  min = 0;
  max = 0;
  calls = 0;
}


CCommand::CCommand( char *newName, char *newCommandID, char *newList, int newMin=0, int newMax=0, int newCalls=0 ) : CTclDataType( newName, newCalls )
{
  commandID = new char[ strlen( newCommandID )+1 ];
  strcpy( commandID, newCommandID );
  list = new char[ strlen( newList )+1 ];
  strcpy( list, newList );
  min = newMin;
  max = newMax;
}

CCommand::CCommand( const CCommand &command )
{
  delete name;
  name = new char[ strlen( command.name )+1 ];
  strcpy( name, command.name );
  delete commandID;
  commandID = new char[ strlen( command.commandID )+1 ];
  strcpy( commandID, command.commandID );
  delete list;
  list = new char[ strlen( command.list )+1 ];
  strcpy( list, command.list );
  min = command.min;
  max = command.max;
  calls = command.calls;
}


CCommand::~CCommand( void )
{
  delete commandID;
  delete list;
}


char *CCommand::CommandID( void )
{
  CopyToBuffer( commandID );
  return Buffer();
}


char *CCommand::List( void )
{
  CopyToBuffer( list );
  return Buffer();
}


int CCommand::Max( void )
{
  return max;
}


int CCommand::Min( void )
{
  return min;
}


void CCommand::SetCommandID( char *newCommandID )
{
  delete commandID;
  commandID = new char[ strlen( newCommandID )+1 ];
  strcpy( commandID, newCommandID );
}


void CCommand::SetList( char *newList )
{
  delete list;
  list = new char[ strlen( newList )+1 ];
  strcpy( list, newList );
}


void CCommand::SetMax( int newMax )
{
  max = newMax;
}


void CCommand::SetMin( int newMin )
{
  min = newMin;
}


int CCommand::operator==( CCommand &command )
{
  //Zum Vergleichen wird nur der Befehlsname verwendet.
  if( !strcmp( name, command.name ) )
    return 1;

  else
    return 0;
}


CCommand &CCommand::operator=( CCommand &command )
{
  if( this == &command )
    return *this;

  delete name;
  name = new char[ strlen( command.name )+1 ];
  strcpy( name, command.name );
  delete commandID;
  commandID = new char[ strlen( command.commandID )+1 ];
  strcpy( commandID, command.commandID );
  delete list;
  list = new char[ strlen( command.list )+1 ];
  strcpy( list, command.list );
  min = command.min;
  max = command.max;
  calls = command.calls;

  return *this;
}


//######################################################################


CVariable::CVariable( void )
{
  name = NULL;
  environment = NULL;
  status = 0;
}


/*
CVariable::CVariable( char *newName, char *newEnvironment, int newStatus=0, int newCalls=0 ) : CTclDataType( newName, newCalls )
{
  environment = new char[ strlen( newEnvironment )+1 ];
  strcpy( environment, newEnvironment );
  status = newStatus;
}
*/


CVariable::CVariable( char *newName, char *newEnvironment, int newStatus=0, int newCalls=0 )
{
  name = new char[ strlen( newName )+1 ];
  strcpy( name, newName );
  environment = new char[ strlen( newEnvironment )+1 ];
  strcpy( environment, newEnvironment );
  status = newStatus;
  calls = newCalls;
}


CVariable::CVariable( const CVariable &variable )
{
  delete name;
  name = new char[ strlen( variable.name )+1 ];
  strcpy( name, variable.name );
  delete environment;
  environment = new char[ strlen( variable.environment )+1 ];
  strcpy( environment, variable.environment );
  status = variable.status;
  calls = variable.calls;
}


CVariable::~CVariable( void )
{
  delete environment;
}


char *CVariable::Environment( void )
{
  CopyToBuffer( environment );
  return Buffer();
}


void CVariable::SetEnvironment( char *newEnvironment )
{
  delete environment;
  environment = new char[ strlen( newEnvironment )+1 ];
  strcpy( environment, newEnvironment );
}


void CVariable::SetStatus( int newStatus )
{
  status = newStatus;
}


int CVariable::Status( void )
{
  return status;
}


int CVariable::operator==( CVariable &variable )
{
  int a, b;
  a = !strcmp( name, variable.name );
  b = !strcmp( environment, variable.environment );
  if( a && b )
    return 1;
  else
    return 0;
}


CVariable &CVariable::operator=( CVariable &variable )
{
  //Bei Selbstzuweisung muss man nichts machen.
  if( &variable == this )
    return *this;

  delete name;
  name = new char[ strlen( variable.name )+1 ];
  strcpy( name, variable.name );

  delete environment;
  environment = new char[ strlen( variable.environment )+1 ];
  strcpy( environment, variable.environment );

  status = variable.status;
  calls = variable.calls;
  return *this;
}


//######################################################################


CHandler::CHandler( void )
{
  //'optionList' wird aus der Datei '.options.tclparse' im HOME-Verzeichniss
  // initialisiert.
  char *home = getenv( "HOME" );

  char fileName[ strlen( home ) + strlen( optionsFileName ) + 2 ];
  strcpy( fileName, home );
  strcat( fileName, "/" );
  strcat( fileName, optionsFileName );

  //Ein Dateiobjekt zum Lesen erzeugen und versuchen, die Datei zu oeffnen.
  ifstream fin( fileName );

  //Eventuellen Fehler abfangen.
  if( fin.fail() )
    {
      display.Message( "options database '", fileName, "' not found, program terminated..." );
      exit( 0 );
    }

  char line[ bufferSize ];
  char name[ bufferSize ];

  //Jetzt die Datei einlesen und im Objekt 'OptionsHandler' speichern.
  while( 1 )
    {
      //Den Namen einlesen.
      fin.getline( line, bufferSize, '\n' );

      //Eventuell wurde das Dateiende erreicht.
      if( fin.eof() )
	break;

      //Den Namen zwischenspeichern.
      strcpy( name, line );

      //Die Liste mit den Schalternamen auslesen.
      fin.getline( line, bufferSize, '\n' );

      //Eventuell wurde das Dateiende erreicht ( sollte nicht vorkommen )
      if( fin.eof() )
	{
	  fin.close();
	  return;
	}

      //Den Namen samt Liste im Listen-Objekt speichern.
      AddOptionList( name, line );
    }

  display.Message( "options database '", fileName, "' loaded" );
  fin.close();
}


CHandler::~CHandler( void )
{
}


int CHandler::IsArray( char *name, char *environment )
{
  CVariable *variable;

  char searchStr[ strlen( name )+2 ];
  strcpy( searchStr, name );
  strcat( searchStr, "(" );
  int endPos = strlen( searchStr );

  char varName[ unitLength ];
  char dummy[ unitLength ];

  char envDummy[ strlen( environment )+1 ];
  strcpy( envDummy, environment );

  //Die aeussere Schleife durchlaeuft die relevanten Umgebungen.
  while( 1 )
    {
      //Die innere Schleife sucht in der jeweiligen Umgebung einen
      //passenden Variablennamen.
      variable = variableList.First();
      while( variable != NULL )
	{
	  strcpy( varName, variable->Name() );
	  strcpy( dummy, varName );

	  //Den entsprechenden Teil des Namens mit dem Suchstring vergleichen.
	  dummy[ endPos ] = '\0';

	  if( !strcmp( dummy, searchStr ) )
	    {
	      //Ist das letzte Zeichen des Namens eine schliessende runde Klammer,
	      //dann sollte die Variable ein Feldelement sein.
	      if( varName[ strlen( varName )-1 ] == ')' )
		return 1;
	    }

	  variable = variableList.Next();
	}

      if( !strcmp( envDummy, " " ) )
	return 0;

      else
	strcpy( envDummy, NextEnvironment( envDummy ) );
    }

  return 0;
}


int CHandler::IsCommand( char *name )
{
  CCommand *command = GetCommand( name );

  if( command == NULL )
    return 0;

  else
    return 1;
}


int CHandler::IsGlobal( char *name )
{
  CVariable *variable = GetVariable( name, " " );

  if( variable == NULL)
    return 0;

  else
    return 1;
}


int CHandler::IsImage( char *name )
{
  CCommand *image = GetCommand( name );

  if( image == NULL )
    return 0;

  if( !strcmp( image->CommandID(), "IMAGE" ) )
    return 1;

  else
    return 0;
}


int CHandler::IsOption( char *option, char *command )
{
  //Die Liste ermitteln.
  CCommand *list = GetOptionList( command );

  if( list == NULL )
    return 0;

  //In der Liste nach der angegebenen Option suchen.
  char dummy[ strlen( list->List() ) ];
  strcpy( dummy, Buffer() );

  return IsElementOf( option, dummy );
}


int CHandler::IsProcedure( char *name )
{
  CCommand *procedure = GetCommand( name );

  if( procedure == NULL )
    return 0;

  if( !strcmp( procedure->CommandID(), "PROCEDURE" ) )
    return 1;

  else
    return 0;
}


int CHandler::IsVariable( char *name, char *environment  )
{
  char dummyEnv[ strlen( environment )+1 ];
  strcpy( dummyEnv, environment );

  //Ein Hilfsobjekt erzeugen.
  CVariable *variable;

  //Pruefen, ob die Variable schon existiert, und zwar in
  //allen moeglichen Umgebungen.
  while( 1 )
    {
      variable = GetVariable( name, dummyEnv );

      //Pruefen, ob die Variable in der aktuellen Umgebung existiert.
      if( variable != NULL )
	return 1;

      //Befindet man sich in der globalen Umgebung, so kann die Schleife
      //verlassen werden.
      if( !strcmp( dummyEnv, " " ) )
	return 0;

      //Die naechsthoehere Umgebung ermitteln.
      strcpy( dummyEnv, NextEnvironment( dummyEnv ) );
    } 
}


int CHandler::IsWidget( char *name )
{
  CCommand *widget = GetCommand( name );

  if( widget == NULL )
    return 0;

  if( !strcmp( widget->CommandID(), "WIDGET" ) )
    return 1;

  else
    return 0;
}


void CHandler::AddCommand( char *name, char *commandID, char *list, int min=0, int max=0, int calls=0 )
{
  if( IsCommand( name ) )
    RemoveCommand( name );

  CCommand command( name, commandID, list, min, max, calls );
  commandList.Insert( command );
}


void CHandler::AddImage( char *name, char *type, int calls=0 )
{
  if( IsCommand( name ) )
    RemoveCommand( name );

  CCommand image( name, "IMAGE", type, 0, 0, calls );
  commandList.Insert( image );
}


void CHandler::AddOptionList( char *name, char *list )
{
  CCommand optList( name, "", list );
  optionList.Insert( optList );
}


void CHandler::AddProcedure( char *name, char *arguments, int min=0, int max=0, int calls=0 )
{
  if( IsCommand( name ) )
    RemoveCommand( name );

  CCommand procedure( name, "PROCEDURE", arguments, min, max, calls );
  commandList.Insert( procedure );
}


int CHandler::AddVariable( char *name, char *environment, int status=0, int calls=0 )
{
  CVariable variable( name, environment, status, calls );

  if( !IsVariable( name, environment ) )
    {
      variableList.Insert( variable );
      return 1;
    }

  else
    {
      //Eventuell den Initialisierungsstatus einer schon vorhandenen
      //Variablen anpassen.
      CVariable *dummy = GetVariable( name, environment );
      if( dummy != NULL )
	{
	  if( !dummy->Status() && status )
	    dummy->SetStatus( 0 );

	  return 1;
	}
	
      return 0;
    }
}


void CHandler::AddWidget( char *name, char *type, int calls=0 )
{
  if( IsCommand( name ) )
    RemoveCommand( name );

  CCommand widget( name, "WIDGET", type, 0, 0, calls );
  commandList.Insert( widget );
}


int CHandler::GetOptionCount( char *option, char *command )
{
  //Die Liste mit den Schalternamen ermitteln.
  CCommand *list = GetOptionList( command );

  if( list == NULL )
    return -1;

  char dummy[ strlen( option )+3 ];
  strcpy( dummy, " " );
  strcat( dummy, option );
  strcat( dummy, " " );

  //Die Stelle mit dem Namen des Schalters in der Liste finden.
  char *result = strstr( list->List(), dummy );

  if( result == NULL )
    return -1;

  //Das naechste Element in der Liste ist die Anzahl der erlaubten Argumente fuer
  //diesen Schalter.

  char *token = strtok( result, " " );
  token = strtok( NULL, " " );

  //String in Zahl umwandeln.
  int count = atoi( token );

  return count;
}


CCommand *CHandler::GetCommand( char *name )
{
  CCommand command( name, "", "" );
  return commandList.Get( command );
}


CCommand *CHandler::GetOptionList( char *name )
{
  CCommand list( name, "", "" );
  return optionList.Get( list );
}


CVariable *CHandler::GetVariable( char *name, char *environment )
{
  char dummyEnv[ strlen( environment )+1 ];
  strcpy( dummyEnv, environment );

  CVariable variable( name, environment );
  CVariable *dummy;

  //Alle Umgebungen aufsteigend durchsuchen.
  while( 1 )
    {
      dummy = variableList.Get( variable );

      //Pruefen, ob die Variable in der aktuellen Umgebung existiert.
      if( dummy != NULL )
	return dummy;

      //Befindet man sich in der globalen Umgebung, so kann die Schleife
      //verlassen werden.
      if( !strcmp( dummyEnv, " " ) )
	return NULL;

      //Die naechsthoehere Umgebung ermitteln.
      strcpy( dummyEnv, NextEnvironment( dummyEnv ) );
      variable.SetEnvironment( dummyEnv );
    } 
}


int CHandler::RemoveCommand( char *name )
{
  CCommand command( name , "", "" );
  return commandList.Remove( command );
}


int CHandler::RemoveVariable( char *name, char *environment )
{
  CVariable variable( name, environment );
  return variableList.Remove( variable );
}


char *CHandler::NextEnvironment( char *envString )
{
  //'strtok' arbeitet dummerweise nicht mit dem uebergebenen Argument 'string'.
  //Deshalb eine neue Variable definieren.
  char environment[ strlen( envString )+1 ];
  strcpy( environment, envString );

  //INFO: Die einzelnen Namen sind durch das Leer-Zeichen getrennt. Die
  //globale Umgebung besitzt die Bezeichnung ' '.

  //Im folgenden wird nicht 'strtok' verwendet, da dies im Programm Probleme
  //verursachen wuerde ( 'strtok' kann nicht verschachtelt aufgerufen werden ).

  //Da das letzte Zeichen von 'environment' kein Leerzeichen ist, ist das
  //letzte Leerzeichen das Ende des gesuchten Strings.

  char *ende = strrchr( environment, " " );

  //1. Die Zeichenkette besteht mindestens aus zwei Teilen.
  if( ende != NULL )
    {
      *ende = '\0';
      CopyToBuffer( environment );
      return Buffer();
    }

  //2. Die Zeichenkette ist leer oder besteht nur aus einem Teil.
  else
    {
      CopyToBuffer( " " );
      return Buffer();
    }
}


char *CHandler::ParentWidget( char *name )
{
  //Hilfszeiger deklarieren.
  char *actual;
  char *next;

  //Den ersten und den zweiten Teil des Umgebungsnamens ermitteln.
  char copy[ strlen( name )+1 ];
  strcpy( copy, name );
  actual = strtok( copy, "." );
  next = strtok( NULL, "." );

  //In die folgende Variable wird der Ergebnisstring kopiert.
  char dummy[ strlen( copy )+1 ];
  strcpy( dummy, "" );

  //In der Schleife den uebergebenen String komplett durchlaufen.
  while( next != NULL )
    {
      //Den Teilnamen an das Ergebnis anhaengen.
      strcat( dummy, "." );
      strcat( dummy, actual );

      actual = next;

      //Den naechsten Teilnamen ermitteln.
      next = strtok( NULL, "." );
    }

  if( !strcmp( dummy, "" ) )
    strcpy( dummy, "." );

  CopyToBuffer( dummy );
  return Buffer();
}


int CHandler::Write( char *fileName )
{
  //Die Informationsdatei zum Schreiben oeffnen. 
  ofstream fout;
  fout.open( fileName );

  if( fout.fail() )
    {
      display.Message( "error opening information file '", fileName, "'" );
      return 0;
    }

  //Zuerst die Befehlsinformationen schreiben.
  //Eine Ueberschrift zur spaeteren Identifizierung beim Laden schreiben.
  fout << commandInfoID << '\n';

  //Den ersten Eintrag ermitteln.
  CCommand *command = commandList.First();
  while( command != NULL )
    {
      fout << command->Name() << '\t';
      fout << command->CommandID() << '\t';
      fout << command->List() << '\t';
      fout << command->Min() << '\t';
      fout << command->Max() << '\t';
      fout << command->Calls() << '\n';

      command = commandList.Next();
    }

  //Jetzt die Variableninformationen schreiben.
  //Eine Ueberschrift zur spaeteren Identifizierung beim Laden schreiben.
  fout << variableInfoID << '\n';

  //Den ersten Eintrag ermitteln.
  CVariable *variable = variableList.First();
  while( variable != NULL )
    {
      fout << variable->Name() << '\t';
      fout << variable->Environment() << '\t';
      fout << variable->Status() << '\t';
      fout << variable->Calls() << '\n';

      variable = variableList.Next();
    }

  fout.close();
  return 1;
}


CHandler &CHandler::operator=( CHandler &handler )
{
  if( &handler == this )
    return *this;

  commandList = handler.commandList;
  variableList = handler.variableList;
  optionList = handler.optionList;

  return *this;
}


//######################################################################



